
function CompSeisB(dir,stat,format);
%function CompSeisB(sc,stat);

% input: stat == string denoting the station
% input: sc == string denoting the folder which has the data files
% stat='BB.s02_m1.hyb';
% sc='TS1_2/';
% Banu Mena, NOV 2007
% --------------------


%% read BB-motions
sc=[dir 'hybrid/'];
fid = fopen([sc,stat],'r'); [sc,stat]
for ii=1:8;
  line=fgetl(fid);
end
data = fscanf(fid,'%g %e %e %e',[4 inf]);  
fclose(fid);
data = data';

t_hf=data(:,1);
ns_hf = data(:,2); ew_hf = data(:,3); ud_hf = data(:,4);
dt_hf=t_hf(2)-t_hf(1);

%% read LF-motions
sc=[dir 'lf_files/'];
index = strfind(stat,'.hyb');
st = stat(4:index-1);

if (format == 'RGF')
   cmp1='.000';  cmp2='.090';  cmp3='.ver';
else 
   cmp1='.xv';  cmp2='.yv';  cmp3='.zv';          
end

st1=[st,cmp1]; st2=[st,cmp2]; st3=[st,cmp3];
  
file1 = fopen([sc,st1],'r');
for ii=1:1;
   line=fgetl(file1);
end

data = fscanf(fid,'%f %f',[2]); dt_lf = data(2); npts = data(1);

t_lf=[0:dt_lf:(npts-1)*dt_lf];
   
ns_lf = fscanf(file1,'%e',[inf]);  
    
file2 = fopen([sc,st2],'r');
for ii=1:2;
   line=fgetl(file2);
end
ew_lf = fscanf(file2,'%e',[inf]);  
  
file3 = fopen([sc,st3],'r');
for ii=1:2;
   line=fgetl(file3);
end
ud_lf = fscanf(file3,'%e',[inf]);  
  
  %% plot time histories
  figure;
  subplot(3,2,1); 
  plot(t_hf,ew_hf); grid on; title('ew'); hold on; plot(t_lf,ew_lf,'r'); %legend('BB','LF');
  subplot(3,2,3);
  plot(t_hf,ns_hf); grid on; title('ns');  hold on; plot(t_lf,ns_lf,'r'); ylabel('Vel (m/s)'); %legend('BB','LF');
  subplot(3,2,5);
  plot(t_hf,ud_hf); grid on; title('ud'); hold on; plot(t_lf,ud_lf,'r'); %legend('BB','LF');
  %legend('LP');
  xlabel('Time (s)');
 
  %% FAS of LF
  [fas_ewlf,f_lf] = sfft(ew_lf,dt_lf);
  [fas_nslf,f_lf] = sfft(ns_lf,dt_lf);
  [fas_udlf,f_lf] = sfft(ud_lf,dt_lf);  
  %FAS of BB
  [fas_ew,f_hf] = sfft(ew_hf,dt_hf);
  [fas_ns,f_hf] = sfft(ns_hf,dt_hf);
  [fas_ud,f_hf] = sfft(ud_hf,dt_hf);
  
  %% plot FAS
  subplot(3,2,2);
  loglog(f_hf,fas_ew); grid on; hold on; loglog(f_lf,fas_ewlf,'r'); %legend('BB','LF');
  subplot(3,2,4);
  loglog(f_hf,fas_ns); grid on; hold on; loglog(f_lf,fas_nslf,'r'); %legend('BB','LF');
  subplot(3,2,6);
  loglog(f_hf,fas_ud); grid on; hold on; loglog(f_lf,fas_udlf,'r'); xlabel('Freq(Hz)'); %legend('BB','LF'); 
  
  
  %% BB Acceleration
  aew=diff(ew_hf)./dt_hf; aew=[aew(1) aew']; aew = aew';
  ans=diff(ns_hf)./dt_hf; ans=[ans(1) ans']; ans = ans';
  aud=diff(ud_hf)./dt_hf; aud=[aud(1) aud']; aud = aud';
  %% FAS of BB Acceleration
  [Abb_ew,ff] = sfft(aew,dt_hf);
  [Abb_ns,ff] = sfft(ans,dt_hf);
  [Abb_ud,ff] = sfft(aud,dt_hf);
  
  %% LF Acceleration
  aew_lf=diff(ew_lf)./dt_lf; aew_lf=[aew_lf(1) aew_lf']; aew_lf = aew_lf';
  ans_lf=diff(ns_lf)./dt_lf; ans_lf=[ans_lf(1) ans_lf']; ans_lf = ans_lf';
  aud_lf=diff(ud_lf)./dt_lf; aud_lf=[aud_lf(1) aud_lf']; aud_lf = aud_lf';
  %% FAS of LF Acceleration
  [Alf_ew,ff_lf] = sfft(aew_lf,dt_lf);
  [Alf_ns,ff_lf] = sfft(ans_lf,dt_lf);
  [Alf_ud,ff_lf] = sfft(aud_lf,dt_lf);
  
  %% Plot Acceleration
  figure;
  subplot(3,2,1); 
  plot(t_hf,aew); grid on; title('ew'); hold on; plot(t_lf,aew_lf,'r'); %legend('BB','LF');
  subplot(3,2,3);
  plot(t_hf,ans); grid on; title('ns');  hold on; plot(t_lf,ans_lf,'r'); ylabel('Acc (m/s^2)'); %legend('BB','LF'); 
  subplot(3,2,5);
  plot(t_hf,aud); grid on; title('ud'); hold on; plot(t_lf,aud_lf,'r'); %legend('BB','LF');
  xlabel('Time (s)');
  %% plot FAS
  subplot(3,2,2);
  loglog(ff,Abb_ew); grid on;  axis([10^-2 10^2 10^-4 10^1]); hold on; loglog(ff_lf,Alf_ew,'r'); %legend('BB','LF');
  subplot(3,2,4);
  loglog(ff,Abb_ns); grid on; hold on; axis([10^-2 10^2 10^-4 10^1]); hold on; loglog(ff_lf,Alf_ns,'r'); %legend('BB','LF');
  subplot(3,2,6);
  loglog(ff,Abb_ud); grid on; hold on; axis([10^-2 10^2 10^-4 10^1]); hold on; loglog(ff_lf,Alf_ud,'r'); xlabel('Freq(Hz)');%legend('BB','LF'); 
