function [am,f,ph] = sfft(y,dt,whole) 
%
%  function [am,f,ph] = sfft(y,dt,whole) 
%  calculates the Fourier Transform of a signal
% 	
%  Input: y  - signal; can be array of many signals of equal
%	       length and sampling, arranged in columns
%	  dt - sampling interval or time axis
%	  whole - optional: flag to get entire spectrum ('y')
%
%  Output: am - amplitude spectrum
%	   f  - frequency axis for pos. frequencies only
%	   ph - phase spectrum
%
%  mmai, 01/29/98
%  --------------

if nargin == 1; dt = 0.01; whole = 'n';
elseif nargin == 2; whole = 'n'; 
end;

m = size(y,1);

if size(y,1) == 1, y = y(:); end
if mod(m,2)  == 1
    m=m-1; y = y(1:m,:);
end


if length(dt) > 1
   dt = dt(2)-dt(1);		% assuming dt is regular spaced t-axis
end

W = 1/dt;			% sampling freq in Hz 
f = (W/m)*linspace(-m/2,m/2,m)';% frequency axis


%%% calculate absolute FFT, scaled by sampling interval 
%%% and shifted symmetrically

FT = (fft(y))*dt;
for i = 1:size(y,2);
  am(:,i) = abs(fftshift(FT(:,i)));
  ph(:,i) = angle(FT(:,i));
end

if whole == 'n';
  AM = am; PH = ph;
  clear am ph;
  f = f(m/2+1:end);
  for i = 1:size(y,2)
    am(:,i) = AM(m/2+1:end,i);
    ph(:,i) = PH(m/2+1:end,i);
  end
end
